@@ -13,17 +13,23 @@ from pyhanko.config import (
 )
 from pyhanko.pdf_utils import misc
 from pyhanko.pdf_utils.config_utils import ConfigurationError
+from pyhanko.pdf_utils.crypt import (
+    SimpleEnvelopeKeyDecrypter,
+    PubKeySecurityHandler, AuthResult,
+)
 
 from pyhanko.sign import signers
 from pyhanko.sign.general import SigningError
 from pyhanko.sign.timestamps import HTTPTimeStamper
 from pyhanko.sign import validation, beid, fields
 from pyhanko.pdf_utils.reader import PdfFileReader
+from pyhanko.pdf_utils.writer import copy_into_new_writer
 from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
 from pyhanko.sign.validation import (
     SignatureValidationError, RevocationInfoValidationType
 )
 from pyhanko.stamp import QRStampStyle, text_stamp_file, qr_stamp_file
+from pyhanko import __version__
 
 __all__ = ['cli']
 
@@ -86,9 +92,11 @@ class Ctx(Enum):
     STAMP_STYLE = auto()
     STAMP_URL = auto()
     NEW_FIELD_SPEC = auto()
+    PREFER_PSS = auto()
 
 
 @click.group()
+@click.version_option(prog_name='pyHanko', version=__version__)
 @click.option('--config',
               help=(
                   'YAML file to load configuration from'
@@ -419,6 +427,8 @@ def lta_update(ctx, infile, validation_context, trust, trust_replace,
 @click.option('--use-pades', help='sign PAdES-style [level B/B-T/B-LT]',
               required=False, default=False, is_flag=True, type=bool,
               show_default=True)
+@click.option('--prefer-pss', is_flag=True, default=False, type=bool,
+              help='prefer RSASSA-PSS to PKCS#1 v1.5 padding, if available')
 @click.option('--with-validation-info', help='embed revocation info',
               required=False, default=False, is_flag=True, type=bool,
               show_default=True)
@@ -435,9 +445,10 @@ def lta_update(ctx, infile, validation_context, trust, trust_replace,
 def addsig(ctx, field, name, reason, location, certify, existing_only,
            timestamp_url, use_pades, with_validation_info,
            validation_context, trust_replace, trust, other_certs,
-           style_name, stamp_url):
+           style_name, stamp_url, prefer_pss):
     ctx.obj[Ctx.EXISTING_ONLY] = existing_only or field is None
     ctx.obj[Ctx.TIMESTAMP_URL] = timestamp_url
+    ctx.obj[Ctx.PREFER_PSS] = prefer_pss
 
     if use_pades:
         subfilter = fields.SigSeedSubFilter.PADES
@@ -466,7 +477,34 @@ def addsig(ctx, field, name, reason, location, certify, existing_only,
     ctx.obj[Ctx.STAMP_URL] = stamp_url
 
 
-def addsig_simple_signer(signer: signers.SimpleSigner, infile, outfile,
+def _open_for_signing(infile_path, signer_cert=None, signer_key=None):
+    from pyhanko.pdf_utils import crypt
+    infile = open(infile_path, 'rb')
+    writer = IncrementalPdfFileWriter(infile)
+
+    # TODO make this an option higher up the tree
+    # TODO mention filename in prompt
+    if writer.prev.encrypted:
+        sh = writer.prev.security_handler
+        if isinstance(sh, crypt.StandardSecurityHandler):
+            pdf_pass = getpass.getpass(
+                prompt='Password for encrypted file \'%s\': ' % infile_path
+            )
+            writer.encrypt(pdf_pass)
+        elif isinstance(sh, crypt.PubKeySecurityHandler) \
+                and signer_key is not None:
+            # attempt to decrypt using signer's credentials
+            cred = SimpleEnvelopeKeyDecrypter(signer_cert, signer_key)
+            writer.encrypt_pubkey(cred)
+        else:
+            raise click.ClickException(
+                "Input file appears to be encrypted, but appropriate "
+                "credentials are not available."
+            )
+    return writer
+
+
+def addsig_simple_signer(signer: signers.SimpleSigner, infile_path, outfile,
                          timestamp_url, signature_meta, existing_fields_only,
                          style, stamp_url, new_field_spec):
     with pyhanko_exception_manager():
@@ -474,15 +512,10 @@ def addsig_simple_signer(signer: signers.SimpleSigner, infile, outfile,
             timestamper = HTTPTimeStamper(timestamp_url)
         else:
             timestamper = None
-        writer = IncrementalPdfFileWriter(infile)
-
-        # TODO make this an option higher up the tree
-        # TODO mention filename in prompt
-        if writer.prev.encrypted:
-            pdf_pass = getpass.getpass(
-                prompt='Password for encrypted file: '
-            ).encode('utf-8')
-            writer.encrypt(pdf_pass)
+        writer = _open_for_signing(
+            infile_path, signer_cert=signer.signing_cert,
+            signer_key=signer.signing_key
+        )
 
         text_params = None
         if stamp_url is not None:
@@ -500,12 +533,12 @@ def addsig_simple_signer(signer: signers.SimpleSigner, infile, outfile,
         outfile.write(buf)
         buf.release()
 
-        infile.close()
+        writer.prev.stream.close()
         outfile.close()
 
 
 @addsig.command(name='pemder', help='read key material from PEM/DER files')
-@click.argument('infile', type=click.File('rb'))
+@click.argument('infile', type=readable_file)
 @click.argument('outfile', type=click.File('wb'))
 @click.option('--key', help='file containing the private key (PEM/DER)', 
               type=readable_file, required=True)
@@ -534,7 +567,7 @@ def addsig_pemder(ctx, infile, outfile, key, cert, chain, passfile):
     
     signer = signers.SimpleSigner.load(
         cert_file=cert, key_file=key, key_passphrase=passphrase,
-        ca_chain_files=chain
+        ca_chain_files=chain, prefer_pss=ctx.obj[Ctx.PREFER_PSS]
     )
     return addsig_simple_signer(
         signer, infile, outfile, timestamp_url=timestamp_url,
@@ -546,7 +579,7 @@ def addsig_pemder(ctx, infile, outfile, key, cert, chain, passfile):
 
 
 @addsig.command(name='pkcs12', help='read key material from a PKCS#12 file')
-@click.argument('infile', type=click.File('rb'))
+@click.argument('infile', type=readable_file)
 @click.argument('outfile', type=click.File('wb'))
 @click.argument('pfx', type=readable_file)
 @click.option('--chain', type=readable_file, multiple=True,
@@ -574,7 +607,8 @@ def addsig_pkcs12(ctx, infile, outfile, pfx, chain, passfile):
         passfile.close()
 
     signer = signers.SimpleSigner.load_pkcs12(
-        pfx_file=pfx, passphrase=passphrase, ca_chain_files=chain
+        pfx_file=pfx, passphrase=passphrase, ca_chain_files=chain,
+        prefer_pss=ctx.obj[Ctx.PREFER_PSS]
     )
     return addsig_simple_signer(
         signer, infile, outfile, timestamp_url=timestamp_url,
@@ -738,3 +772,151 @@ def stamp(ctx, infile, outfile, x, y, style_name, page, stamp_url):
             text_stamp_file(
                 infile, outfile, stamp_style, dest_page=page_ix, x=x, y=y
             )
+
+
+@cli.command(help='encrypt PDF files (AES-256 only)', name='encrypt')
+@click.argument('infile', type=readable_file)
+@click.argument('outfile', type=click.Path(writable=True, dir_okay=False))
+@click.option(
+    '--password', help='password to encrypt the file with', required=False,
+    type=str
+)
+@click.option(
+    '--recipient', required=False, multiple=True,
+    help='certificate(s) corresponding to entities that '
+         'can decrypt the output file',
+    type=click.Path(readable=True, dir_okay=False)
+)
+def encrypt_file(infile, outfile, password, recipient):
+    if bool(password) == bool(recipient):
+        raise click.ClickException(
+            "Specify either a password or a list of recipients."
+        )
+
+    recipient_certs = None
+    if recipient:
+        recipient_certs = list(
+            signers.load_certs_from_pemder(cert_files=recipient)
+        )
+
+    with pyhanko_exception_manager():
+        with open(infile, 'rb') as inf:
+            r = PdfFileReader(inf)
+            w = copy_into_new_writer(r)
+
+            if recipient_certs:
+                w.encrypt_pubkey(recipient_certs)
+            else:
+                w.encrypt(owner_pass=password)
+
+            with open(outfile, 'wb') as outf:
+                w.write(outf)
+
+
+@cli.group(help='decrypt PDF files (any standard PDF encryption scheme)',
+           name='decrypt')
+def decrypt():
+    pass
+
+
+decrypt_force_flag = click.option(
+    '--force', help='ignore access restrictions (use at your own risk)',
+    required=False, type=bool, is_flag=True, default=False
+)
+
+
+@decrypt.command(help='decrypt using password', name='password')
+@click.argument('infile', type=readable_file)
+@click.argument('outfile', type=click.Path(writable=True, dir_okay=False))
+@click.option(
+    '--password', help='password to decrypt the file with', required=False,
+    type=str
+)
+@decrypt_force_flag
+def decrypt_with_password(infile, outfile, password, force):
+    with pyhanko_exception_manager():
+        with open(infile, 'rb') as inf:
+            r = PdfFileReader(inf)
+            if r.security_handler is None:
+                raise click.ClickException("File is not encrypted.")
+            if not password:
+                password = getpass.getpass(prompt='File password: ')
+            auth_result = r.decrypt(password)
+            if auth_result == AuthResult.USER and not force:
+                raise click.ClickException(
+                    "Password specified was the user password, not "
+                    "the owner password. Pass --force to decrypt the "
+                    "file anyway."
+                )
+            elif auth_result == AuthResult.FAILED:
+                raise click.ClickException("Password didn't match.")
+            w = copy_into_new_writer(r)
+            with open(outfile, 'wb') as outf:
+                w.write(outf)
+
+
+@decrypt.command(help='decrypt using private key (PEM/DER)', name='pemder')
+@click.argument('infile', type=readable_file)
+@click.argument('outfile', type=click.Path(writable=True, dir_okay=False))
+@click.option('--key', type=readable_file, required=True,
+              help='file containing the recipient\'s private key (PEM/DER)')
+@click.option('--cert', help='file containing the recipient\'s certificate '
+                             '(PEM/DER)', type=readable_file, required=True)
+@click.option('--passfile', required=False, type=click.File('rb'),
+              help='file containing the passphrase for the private key',
+              show_default='stdin')
+@decrypt_force_flag
+def decrypt_with_pemder(infile, outfile, key, cert, passfile, force):
+    if passfile is None:
+        passphrase = getpass.getpass(prompt='Key passphrase: ').encode('utf-8')
+    else:
+        passphrase = passfile.read()
+        passfile.close()
+    sedk = SimpleEnvelopeKeyDecrypter.load(key, cert, key_passphrase=passphrase)
+
+    _decrypt_pubkey(sedk, infile, outfile, force)
+
+
+def _decrypt_pubkey(sedk: SimpleEnvelopeKeyDecrypter, infile, outfile, force):
+    with pyhanko_exception_manager():
+        with open(infile, 'rb') as inf:
+            r = PdfFileReader(inf)
+            if r.security_handler is None:
+                raise click.ClickException("File is not encrypted.")
+            if not isinstance(r.security_handler, PubKeySecurityHandler):
+                raise click.ClickException(
+                    "File was not encrypted with a public-key security handler."
+                )
+            auth_result = r.decrypt_pubkey(sedk)
+            if auth_result == AuthResult.USER:
+                # TODO read 2nd bit of perms in CMS enveloped data
+                #  is the one indicating that change of encryption is OK
+                if not force:
+                    raise click.ClickException(
+                        "Change of encryption is typically not allowed with "
+                        "user access. Pass --force to decrypt the file anyway."
+                    )
+            elif auth_result == AuthResult.FAILED:
+                raise click.ClickException("Failed to decrypt the file.")
+            w = copy_into_new_writer(r)
+            with open(outfile, 'wb') as outf:
+                w.write(outf)
+
+
+@decrypt.command(help='decrypt using private key (PKCS#12)', name='pkcs12')
+@click.argument('infile', type=readable_file)
+@click.argument('outfile', type=click.Path(writable=True, dir_okay=False))
+@click.argument('pfx', type=readable_file)
+@click.option('--passfile', required=False, type=click.File('rb'),
+              help='file containing the passphrase for the PKCS#12 file',
+              show_default='stdin')
+@decrypt_force_flag
+def decrypt_with_pkcs12(infile, outfile, pfx, passfile, force):
+    if passfile is None:
+        passphrase = getpass.getpass(prompt='Key passphrase: ').encode('utf-8')
+    else:
+        passphrase = passfile.read()
+        passfile.close()
+    sedk = SimpleEnvelopeKeyDecrypter.load_pkcs12(pfx, passphrase=passphrase)
+
+    _decrypt_pubkey(sedk, infile, outfile, force)
