@@ -13,12 +13,11 @@ from typing import Iterator, Tuple, Optional, Union, Callable, Any
 from dataclasses import dataclass, field
 
 from .misc import (
-    read_non_whitespace, skip_over_comment, read_until_regex,
+    read_non_whitespace, skip_over_comment, read_until_regex
 )
 from .misc import PdfStreamError, PdfReadError, IndirectObjectExpected
 import logging
 from . import filters
-from .crypt import rc4_encrypt
 import decimal
 import codecs
 
@@ -268,18 +267,18 @@ class PdfObject:
         """
         return self
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None,
+                        container_ref: Reference = None):
         """
         Abstract method to render this object to an output stream.
 
         :param stream:
             An output stream.
-        :param encryption_key:
-            Key to encrypt the object with.
+        :param container_ref:
+            Local encryption key.
+        :param handler:
+            Security handler
         """
-        # TODO put a reference to the encryption docs here (as soon as I have
-        #  those), since it's worth stressing that we only support the crappy
-        #  RC4 encryption scheme of yore.
         raise NotImplementedError
 
 
@@ -289,7 +288,7 @@ class NullObject(PdfObject):
     All instances are treated as equal and falsy.
     """
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         stream.write(b"null")
 
     @staticmethod
@@ -315,7 +314,7 @@ class BooleanObject(PdfObject):
     def __init__(self, value):
         self.value = value
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         if self.value:
             stream.write(b"true")
         else:
@@ -375,11 +374,13 @@ class ArrayObject(list, PdfObject):
         val = list.__getitem__(self, index)
         return _deproxy_decrypt(val) if decrypt else val
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         stream.write(b"[")
         for data in self:
             stream.write(b" ")
-            data.write_to_stream(stream, encryption_key)
+            data.write_to_stream(
+                stream, handler=handler, container_ref=container_ref
+            )
         stream.write(b" ]")
 
     @staticmethod
@@ -467,7 +468,7 @@ class IndirectObject(PdfObject, Dereferenceable):
     def __ne__(self, other):
         return not self.__eq__(other)
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         stream.write(b"%d %d R" % (self.idnum, self.generation))
 
     @staticmethod
@@ -529,7 +530,7 @@ class FloatObject(decimal.Decimal, PdfObject):
         """
         return float(self)
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         stream.write(repr(self).encode('ascii'))
 
 
@@ -554,7 +555,7 @@ class NumberObject(int, PdfObject):
         """
         return int(self)
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         stream.write(repr(self).encode('ascii'))
 
     @staticmethod
@@ -714,10 +715,14 @@ class ByteStringObject(bytes, PdfObject):
     For compatibility with :attr:`.TextStringObject.original_bytes`
     """
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         bytearr = self
-        if encryption_key:
-            bytearr = rc4_encrypt(encryption_key, bytearr)
+        if handler is not None and container_ref is not None:
+            cf = handler.get_string_filter()
+            local_key = cf.derive_object_key(
+                container_ref.idnum, container_ref.generation
+            )
+            bytearr = cf.encrypt(local_key, bytearr)
         stream.write(b"<")
         stream.write(binascii.hexlify(bytearr))
         stream.write(b">")
@@ -759,7 +764,7 @@ class TextStringObject(str, PdfObject):
         else:
             raise Exception("no information about original bytes")
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         # Try to write the string out as a PDFDocEncoding encoded string.  It's
         # nicer to look at in the PDF file.  Sadly, we take a performance hit
         # here for trying...
@@ -768,10 +773,14 @@ class TextStringObject(str, PdfObject):
             bytearr = encode_pdfdocencoding(self)
         except UnicodeEncodeError:
             bytearr = codecs.BOM_UTF16_BE + self.encode("utf-16be")
-        if encryption_key:
-            bytearr = rc4_encrypt(encryption_key, bytearr)
+        if handler is not None and container_ref is not None:
+            cf = handler.get_string_filter()
+            local_key = cf.derive_object_key(
+                container_ref.idnum, container_ref.generation
+            )
+            bytearr = cf.encrypt(local_key, bytearr)
             obj = ByteStringObject(bytearr)
-            obj.write_to_stream(stream, None)
+            obj.write_to_stream(stream)
         else:
             stream.write(b"(")
             for c in bytearr:
@@ -792,10 +801,8 @@ class NameObject(str, PdfObject):
 
     DELIMITER_PATTERN = re.compile(r"\s+|[\(\)<>\[\]{}/%]".encode('ascii'))
 
-    def write_to_stream(self, stream, encryption_key):
-        # TODO look up the correct encoding to use in the spec
-        #  (although these will be alphanumeric in 99% percent of cases)
-        stream.write(self.encode('utf-8'))
+    def write_to_stream(self, stream, handler=None, container_ref=None):
+        stream.write(encode_pdfdocencoding(self))
 
     @staticmethod
     def read_from_stream(stream, strict=True):
@@ -899,12 +906,12 @@ class DictionaryObject(dict, PdfObject):
             raise KeyError
         return value
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         stream.write(b"<<\n")
         for key, value in list(self.items()):
-            key.write_to_stream(stream, encryption_key)
+            key.write_to_stream(stream, handler, container_ref)
             stream.write(b" ")
-            value.write_to_stream(stream, encryption_key)
+            value.write_to_stream(stream, handler, container_ref)
             stream.write(b"\n")
         stream.write(b">>")
 
@@ -1020,7 +1027,8 @@ class StreamObject(DictionaryObject):
     The latter will be overwritten as necessary.
     """
 
-    def __init__(self, dict_data=None, stream_data=None, encoded_data=None):
+    def __init__(self, dict_data=None, stream_data=None, encoded_data=None,
+                 handler=None):
         """Initialise a stream with dictionary data and stream data
         (either encoded or decoded). If both `stream_data` and `encoded_data`
         are provided, the caller is responsible for making sure that both are
@@ -1032,10 +1040,35 @@ class StreamObject(DictionaryObject):
             The (unencoded) stream data.
         :param encoded_data:
             The encoded stream data.
+        :param handler:
+            A reference to the currently active
+            :class:`.pyhanko.pdf_utils.crypt.SecurityHandler`.
+            This is only necessary if the stream requires crypt filters.
         """
         super().__init__(dict_data)
         self._data = stream_data
         self._encoded_data = encoded_data
+        self._handler = handler
+
+    @property
+    def _has_crypt_filter(self) -> bool:
+        return '/Crypt' in (name for name, _ in self._filters())
+
+    def add_crypt_filter(self, name=NameObject('/Identity'),
+                         params=None, handler=None):
+        if handler is not None:
+            self._handler = handler
+
+        if name not in self._handler.crypt_filter_config:
+            raise PdfStreamError(
+                f"The crypt filter {name} is not known to the security handler."
+            )
+        params = params or DictionaryObject()
+        params['/Type'] = pdf_name('/CryptFilterDecodeParms')
+        params['/Name'] = name
+        self.apply_filter(
+            pdf_name('/Crypt'), params=params, allow_duplicates=True
+        )
 
     def _filters(self) -> Iterator[Tuple[str, Optional[dict]]]:
         try:
@@ -1060,18 +1093,38 @@ class StreamObject(DictionaryObject):
                 lendiff = len(filter_arr) - len(decode_params)
                 # this should be zero, but let's be lenient
                 if lendiff > 0:
-                    decode_params += [None] * lendiff
+                    decode_params += [NullObject()] * lendiff
         except KeyError:
-            decode_params = [None] * len(filter_arr)
+            decode_params = [NullObject()] * len(filter_arr)
 
-        yield from zip(filter_arr, decode_params)
+        # make sure to deal with resolving decrypted object proxies by
+        # calling get_object()
+        yield from zip(
+            filter_arr, (param_set.get_object() for param_set in decode_params)
+        )
 
-    def _stream_decoders(self) -> Iterator[Tuple[filters.Decoder, dict]]:
+    def _stream_decoders(self):
         for filter_type, params in self._filters():
             try:
                 if params is None or isinstance(params, NullObject):
                     params = {}
-                yield filters.DECODERS[filter_type], params
+                if filter_type == '/Crypt':
+                    # crypt filters get special treatment
+                    # if we're dealing with the identity filter, just move on
+                    if params.get('/Name', '/Identity') == '/Identity':
+                        continue
+                    # if it's another one, we need a reference to the security
+                    # handler
+                    sh = self._handler
+                    if sh is None:
+                        raise PdfStreamError(
+                            "PDF streams require a security handler to use "
+                            "explicit /Crypt filters."
+                        )
+                    decoder = filters.CryptFilterDecoder(sh)
+                else:
+                    decoder = filters.get_generic_decoder(filter_type)
+                yield decoder, params
             except KeyError:
                 raise NotImplementedError(
                     "Filters of type %s are not supported." % filter_type
@@ -1180,12 +1233,16 @@ class StreamObject(DictionaryObject):
                     return
 
             # prepend the new filter (order is important!)
-            self[pdf_name('/Filter')] = ArrayObject((pdf_name,) + filter_names)
+            self[pdf_name('/Filter')] = ArrayObject(
+                (filter_name,) + filter_names
+            )
 
             if params or any(param_sets):
-                self[pdf_name('/DecodeParms')] = [params or NullObject()] + [
-                    param_set or NullObject() for param_set in param_sets
-                ]
+                def _params():
+                    yield params or NullObject()
+                    for param_set in param_sets:
+                        yield param_set or NullObject()
+                self[pdf_name('/DecodeParms')] = ArrayObject(_params())
         self._encoded_data = None
         self._data = data
 
@@ -1198,15 +1255,20 @@ class StreamObject(DictionaryObject):
         """
         self.apply_filter(pdf_name('/FlateDecode'), allow_duplicates=None)
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         data = self.encoded_data
+        if handler is not None and container_ref is not None and \
+                not self._has_crypt_filter:
+            cf = handler.get_stream_filter()
+            local_key = cf.derive_object_key(
+                container_ref.idnum, container_ref.generation
+            )
+            data = cf.encrypt(local_key, data)
         self[NameObject("/Length")] = NumberObject(len(data))
         # write the dictionary
-        super().write_to_stream(stream, encryption_key)
+        super().write_to_stream(stream, handler, container_ref)
         del self["/Length"]
         stream.write(b"\nstream\n")
-        if encryption_key:
-            data = rc4_encrypt(encryption_key, data)
         stream.write(data)
         stream.write(b"\nendstream")
 
@@ -1280,9 +1342,9 @@ pdf_name = NameObject
 PROXYABLE = (TextStringObject, ByteStringObject, DictionaryObject, ArrayObject)
 
 
-def proxy_encrypted_obj(encrypted_obj, key):
+def proxy_encrypted_obj(encrypted_obj, handler):
     if isinstance(encrypted_obj, PROXYABLE):
-        return DecryptedObjectProxy(encrypted_obj, key)
+        return DecryptedObjectProxy(encrypted_obj, handler)
     else:
         return encrypted_obj
 
@@ -1290,10 +1352,10 @@ def proxy_encrypted_obj(encrypted_obj, key):
 # TODO support 2.0-style encryption
 class DecryptedObjectProxy(PdfObject):
 
-    def __init__(self, raw_object: PdfObject, key):
+    def __init__(self, raw_object: PdfObject, handler):
         self.raw_object = raw_object
-        self.key = key
         self._decrypted = None
+        self.handler = handler
 
     @property
     def decrypted(self):
@@ -1301,39 +1363,68 @@ class DecryptedObjectProxy(PdfObject):
         if decrypted is not None:
             return decrypted
 
+        from .crypt import SecurityHandler
+
         obj = self.raw_object
-        key = self.key
+        handler: SecurityHandler = self.handler
+        container_ref = obj.container_ref
+        if not isinstance(container_ref, Reference):
+            raise ValueError(
+                "Proxyable objects must have a container ref pointing to a "
+                f"numbered object, not '{container_ref}'."
+            )  # pragma: nocover
         if isinstance(obj, ByteStringObject) or \
                 isinstance(obj, TextStringObject):
-            decrypted = pdf_string(rc4_encrypt(key, obj.original_bytes))
+            cf = handler.get_string_filter()
+            local_key = cf.derive_object_key(
+                container_ref.idnum, container_ref.generation
+            )
+            decrypted = pdf_string(cf.decrypt(local_key, obj.original_bytes))
         elif isinstance(obj, DictionaryObject):
             decrypted_entries = {
-                dictkey: proxy_encrypted_obj(value, key)
+                dictkey: proxy_encrypted_obj(value, handler)
                 for dictkey, value in obj.items()
             }
             if isinstance(obj, StreamObject):
-                # TODO add tests for this specific use case!
+                # can't deal with crypt filters here
+                if obj._has_crypt_filter:
+                    # in this case, dealing with encryption is delegated
+                    # to the stream decoding process, so just pretend the data
+                    # is decrypted.
+                    # We pass a reference to the security handler below,
+                    # which is sufficient to take care of /Crypt filters
+                    # in the stream.
+                    decrypted_data = obj.encoded_data
+                else:
+                    cf = handler.get_string_filter()
+                    local_key = cf.derive_object_key(
+                        container_ref.idnum, container_ref.generation
+                    )
+                    decrypted_data = cf.decrypt(local_key, obj.encoded_data)
+
                 decrypted = StreamObject(
-                    decrypted_entries,
-                    encoded_data=rc4_encrypt(key, obj.encoded_data)
+                    decrypted_entries, encoded_data=decrypted_data,
+                    handler=handler
                 )
             else:
                 decrypted = DictionaryObject(decrypted_entries)
         elif isinstance(obj, ArrayObject):
-            decrypted_map = map(lambda v: proxy_encrypted_obj(v, key), obj)
+            decrypted_map = map(
+                lambda v: proxy_encrypted_obj(v, handler), obj
+            )
             decrypted = ArrayObject(decrypted_map)
-        else:
+        else:  # pragma: nocover
             raise TypeError(f'Object of type {type(obj)} is not proxyable.')
         decrypted.container_ref = obj.container_ref
         self._decrypted = decrypted
         return decrypted
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         # maybe the encryption key for this object changed (due to it being
         # included as part of a larger object or somesuch, without proper
         # dereferencing), so to avoid unexpected shenanigans, let's start from
         # scratch.
-        self.decrypted.write_to_stream(stream, encryption_key)
+        self.decrypted.write_to_stream(stream, handler, container_ref)
 
     def get_object(self):
         return self.decrypted.get_object()
