@@ -6,6 +6,7 @@ Taken from PyPDF2 with modifications and additions, see
 
 
 import os
+from dataclasses import dataclass
 from enum import Enum
 
 __all__ = [
@@ -244,3 +245,34 @@ def chunked_digest(temp_buffer: bytearray, stream, md, max_read=None):
         else:
             to_feed = read_buffer
         md.update(to_feed)
+
+
+class Singleton(type):
+
+    def __new__(mcs, name, bases, dct):
+        cls = type.__new__(mcs, name, bases, dct)
+        instance = type.__call__(cls)
+        cls.__new__ = lambda _: instance
+        return cls
+
+
+@dataclass(frozen=True)
+class ConsList:
+    head: object
+    tail: 'ConsList' = None
+
+    @staticmethod
+    def empty() -> 'ConsList':
+        return ConsList(head=None)
+
+    def __iter__(self):
+        cur = self
+        while cur.head is not None:
+            yield cur.head
+            cur = cur.tail
+
+    def cons(self, head):
+        return ConsList(head, self)
+
+    def __repr__(self):  # pragma: nocover
+        return f"ConsList({list(reversed(list(self)))})"
