@@ -16,11 +16,14 @@ import re
 from collections import defaultdict
 from io import BytesIO
 from itertools import chain
-from typing import Set, List
+from typing import Set, List, Optional, Union, Tuple
 
 from . import generic, misc
 from .misc import PdfReadError
-from .crypt import _alg33_1, _alg34, _alg35, derive_key, rc4_encrypt
+from .crypt import (
+    SecurityHandler, StandardSecurityHandler,
+    EnvelopeKeyDecrypter, PubKeySecurityHandler,
+)
 
 import logging
 
@@ -502,12 +505,13 @@ class TrailerDictionary(generic.PdfObject):
     def flatten(self, revision=None) -> generic.DictionaryObject:
         relevant_revisions = self._trailer_revisions
         if revision is not None:
-            relevant_revisions = relevant_revisions[:revision]
+            relevant_revisions = relevant_revisions[-revision-1:]
         trailer = generic.DictionaryObject({
             k: v for revision in reversed(relevant_revisions)
             for k, v in revision.items()
         })
-        trailer.update(self._new_changes)
+        if revision is None:
+            trailer.update(self._new_changes)
 
         # ensure that the trailer isn't polluted using stream
         # compression / XRef parameters
@@ -533,8 +537,8 @@ class TrailerDictionary(generic.PdfObject):
     def items(self):
         return self.flatten().items()
 
-    def write_to_stream(self, stream, encryption_key):
-        return self.flatten().write_to_stream(stream, encryption_key)
+    def write_to_stream(self, stream, handler=None, container_ref=None):
+        return self.flatten().write_to_stream(stream, handler, container_ref)
 
 
 class PdfFileReader(PdfHandler):
@@ -555,6 +559,7 @@ class PdfFileReader(PdfHandler):
             problems and also causes some correctable problems to be fatal.
             Defaults to ``True``.
         """
+        self.security_handler: Optional[SecurityHandler] = None
         self.strict = strict
         self.resolved_objects = {}
         self.input_version = None
@@ -580,6 +585,10 @@ class PdfFileReader(PdfHandler):
         except KeyError:
             pass
 
+        encrypt_dict = self._get_encryption_params()
+        if encrypt_dict is not None:
+            self.security_handler = SecurityHandler.build(encrypt_dict)
+
         self._embedded_signatures = None
 
     def _get_object_from_stream(self, idnum, stmnum, idx):
@@ -631,8 +640,11 @@ class PdfFileReader(PdfHandler):
             raise PdfReadError("This is a fatal error in strict mode.")
         return generic.NullObject()
 
-    def _get_encryption_params(self):
-        encrypt_ref = self.trailer.raw_get('/Encrypt')
+    def _get_encryption_params(self) -> Optional[generic.DictionaryObject]:
+        try:
+            encrypt_ref = self.trailer.raw_get('/Encrypt')
+        except KeyError:
+            return
         if isinstance(encrypt_ref, generic.IndirectObject):
             return self.get_object(encrypt_ref.reference, never_decrypt=True)
         else:
@@ -646,6 +658,11 @@ class PdfFileReader(PdfHandler):
     def root_ref(self) -> generic.Reference:
         return self.trailer.raw_get('/Root', decrypt=False).reference
 
+    @property
+    def document_id(self) -> Tuple[bytes, bytes]:
+        id_arr = self.trailer['/ID']
+        return id_arr[0].original_bytes, id_arr[1].original_bytes
+
     def get_historical_root(self, revision: int):
         """
         Get the document catalog for a specific revision.
@@ -755,14 +772,10 @@ class PdfFileReader(PdfHandler):
 
             # override encryption is used for the /Encrypt dictionary
             if not never_decrypt and self.encrypted:
-                try:
-                    shared_key = self._decryption_key
-                except AttributeError:
-                    raise PdfReadError("file has not been decrypted")
-                key = derive_key(shared_key, ref.idnum, ref.generation)
+                sh: SecurityHandler = self.security_handler
                 # make sure the object that lands in the cache is always
                 # a proxy object
-                retval = generic.proxy_encrypted_obj(retval, key)
+                retval = generic.proxy_encrypted_obj(retval, sh)
             return retval
 
     def cache_get_indirect_object(self, generation, idnum):
@@ -893,99 +906,77 @@ class PdfFileReader(PdfHandler):
         self.last_startxref = process_data_at_eof(stream)
         self._read_xrefs()
 
-    # TODO: use sane return values (leftover from PyPDF2)
-    # TODO: support AES
-    def decrypt(self, password: bytes) -> int:
+    def decrypt(self, password: Union[str, bytes]):
         """
-        When using an encrypted PDF file with the PDF legacy RC4-based
-        encryption handler, this function will allow the file to be decrypted.
+        When using an encrypted PDF file with the standard PDF encryption
+        handler, this function will allow the file to be decrypted.
         It checks the given password against the document's user password and
         owner password, and then stores the resulting decryption key if either
         password is correct.
 
-        Supplying either user or owner password will work.
+        Both legacy encryption schemes and PDF 2.0 encryption (based on AES-256)
+        are supported.
 
         .. danger::
-            One should also be aware that the encryption scheme implemented here
-            is (very) weak, and we only support it for compatibility reasons.
-            Under no circumstances should it still be used to encrypt new files.
+            Supplying either user or owner password will work.
+            Cryptographically, both allow the decryption key to be computed,
+            but processors are expected to adhere to the ``/P`` flags in the
+            encryption dictionary when accessing a file with the user password.
+            Currently, pyHanko does not enforce these restrictions, but it
+            may in the future.
+
+        .. danger::
+            One should also be aware that the legacy encryption schemes used
+            prior to PDF 2.0 are (very) weak, and we only support them for
+            compatibility reasons.
+            Under no circumstances should these still be used to encrypt new
+            files.
 
         :param password: The password to match.
-        :return: ``0`` if the password failed, ``1`` if the password matched the
-            user password, and ``2`` if the password matched the owner password.
-        :raises NotImplementedError:
-            Raised if the document uses an unsupported encryption method.
         """
+        sh = self.security_handler
+        if not isinstance(sh, StandardSecurityHandler):
+            raise misc.PdfReadError(
+                f"Security handler is of type '{type(sh)}', "
+                f"not StandardSecurityHandler"
+            )  # pragma: nocover
 
-        return self._decrypt(password)
+        return sh.authenticate(password, id1=self.document_id[0])
 
-    def _decrypt(self, password):
-        encrypt = self._get_encryption_params()
-        if encrypt['/Filter'] != '/Standard':
-            raise NotImplementedError(
-                "only Standard PDF encryption handler is available"
-            )
-        if not (encrypt['/V'] in (1, 2)):
-            raise NotImplementedError(
-                "only algorithm code 1 and 2 are supported"
-            )
-        user_password, key = self._auth_user_password(password)
-        if user_password:
-            self._decryption_key = key
-            return 1
-        else:
-            rev = encrypt['/R'].get_object()
-            if rev == 2:
-                keylen = 5
-            else:
-                keylen = encrypt['/Length'].get_object() // 8
-            key = _alg33_1(password, rev, keylen)
-            owner_token = encrypt["/O"].get_object()
-            if rev == 2:
-                userpass = rc4_encrypt(key, owner_token)
-            else:
-                val = owner_token
-                for i in range(19, -1, -1):
-                    new_key = bytes(b ^ i for b in key)
-                    val = rc4_encrypt(new_key, val)
-                userpass = val
-            owner_password, key = self._auth_user_password(userpass)
-            if owner_password:
-                self._decryption_key = key
-                return 2
-        return 0
+    def decrypt_pubkey(self, credential: EnvelopeKeyDecrypter):
+        """
+        Decrypt a PDF file encrypted using public-key encryption by providing
+        a credential representing the private key of one of the recipients.
 
-    def _auth_user_password(self, password):
-        encrypt = self._get_encryption_params()
-        rev = encrypt['/R'].get_object()
-        owner_entry = encrypt['/O'].get_object()
-        p_entry = encrypt['/P'].get_object()
-        id_entry = self.trailer['/ID'].get_object()
-        id1_entry = id_entry[0].get_object()
-        user_token = encrypt['/U'].get_object().original_bytes
-        if rev == 2:
-            user_tok_supplied, key = _alg34(
-                password, owner_entry, p_entry, id1_entry
-            )
-        elif rev >= 3:
-            encrypt_meta = encrypt.get(
-                "/EncryptMetadata", generic.BooleanObject(False)
-            ).get_object()
-            user_tok_supplied, key = _alg35(
-                password, rev, encrypt["/Length"].get_object() // 8,
-                owner_entry, p_entry, id1_entry, encrypt_meta)
-            user_tok_supplied = user_tok_supplied[:16]
-            user_token = user_token[:16]
-        else:
-            raise NotImplementedError
-        return user_tok_supplied == user_token, key
+        .. danger::
+            The same caveats as in :meth:`.decrypt` w.r.t. permission handling
+            apply to this method.
+
+        .. danger::
+            The robustness of the public key cipher being used is not the only
+            factor in the security of public-key encryption in PDF.
+            The standard still permits weak schemes to encrypt the actual file
+            data and file keys.
+            PyHanko uses sane defaults everywhere, but other software may not.
+
+        :param credential:
+            The :class:`.EnvelopeKeyDecrypter` handling the recipient's
+            private key.
+        """
+        sh = self.security_handler
+        if not isinstance(sh, PubKeySecurityHandler):
+            raise misc.PdfReadError(
+                f"Security handler is of type '{type(sh)}', "
+                f"not PubKeySecurityHandler"
+            )  # pragma: nocover
+        return sh.authenticate(credential)
 
     @property
     def encrypted(self):
         """
         :return: ``True`` if a document is encrypted, ``False`` otherwise.
         """
-        return "/Encrypt" in self.trailer
+        return self.security_handler is not None
 
     def get_historical_resolver(self, revision: int) -> 'HistoricalResolver':
         """
@@ -1036,6 +1027,55 @@ def convert_to_int(d, size):
         return sum(digit ** (size - ix - 1) for ix, digit in enumerate(d))
 
 
+class RawPdfPath:
+    """
+    Class to model raw paths in a file.
+    """
+
+    def __init__(self, *path: Union[str, int]):
+        self.path = path
+
+    def __len__(self):
+        return len(self.path)
+
+    def __iter__(self):
+        return iter(self.path)
+
+    def _tag(self):
+        # should give better hashing results
+        return tuple(map(lambda x: (isinstance(x, int), x), self.path))
+
+    def __hash__(self):
+        return hash(self._tag())
+
+    def __eq__(self, other):
+        return (
+            isinstance(other, RawPdfPath)
+            and (self is other or self._tag() == other._tag())
+        )
+
+    @staticmethod
+    def _fmt_node(node):
+        if isinstance(node, int):
+            return '[%d]' % node
+        else:
+            return '.' + node[1:]
+
+    def __add__(self, other):
+        if isinstance(other, RawPdfPath):
+            return RawPdfPath(*self.path, *other.path)
+        elif isinstance(other, (int, str)):
+            return RawPdfPath(*self.path, other)
+        else:  # pragma: nocover
+            raise TypeError
+
+    def __str__(self):
+        return ''.join(map(RawPdfPath._fmt_node, self.path))
+
+    def __repr__(self):  # pragma: nocover
+        return f"PathInRevision('{str(self)}')"
+
+
 class HistoricalResolver(PdfHandler):
     """
     :class:`~.rw_common.PdfHandler` implementation that provides a view
@@ -1055,11 +1095,17 @@ class HistoricalResolver(PdfHandler):
         selected revision.
     """
 
+    @property
+    def document_id(self) -> Tuple[bytes, bytes]:
+        id_arr = self._trailer['/ID']
+        return id_arr[0].original_bytes, id_arr[1].original_bytes
+
     def __init__(self, reader: PdfFileReader, revision):
         self.cache = {}
         self.reader = reader
         self.revision = revision
         self._trailer = self.reader.trailer.flatten(self.revision)
+        self._indirect_object_access_cache = None
 
     @property
     def trailer_view(self) -> generic.DictionaryObject:
@@ -1194,3 +1240,88 @@ class HistoricalResolver(PdfHandler):
         elif isinstance(obj, generic.ArrayObject):
             for v in obj:
                 self._collect_indirect_references(v, seen, since_revision)
+
+    def _get_usages_of_ref(self, ref: generic.Reference) \
+            -> Optional[Set[RawPdfPath]]:
+        cache = self._indirect_object_access_cache or {}
+        return cache.get(ref, None)
+
+    def _load_reverse_xref_cache(self):
+        if self._indirect_object_access_cache is not None:
+            return
+
+        collected = defaultdict(set)
+
+        # internally, _compute_paths_to_refs works with singly linked lists
+        # to avoid having to create & destroy lots of list objects
+        # We flatten everything when we're done
+        def _compute_paths_to_refs(obj, cur_path: misc.ConsList,
+                                   seen_in_path: misc.ConsList, *,
+                                   is_page_tree, page_tree_objs):
+
+            # optimisation: page tree gets special treatment
+            # to prevent unnecessary paths from being generated when the
+            # tree is entered from the outside (e.g. by following /P on a form
+            # field/widget annotation)
+            # TODO what about the structure tree?
+            # Can we deal with this more systematically? The problem is that
+            # there's no a priori way to state which path from an object
+            # to the trailer is the "canonical" one. For page objects
+            #  things are a bit more clear-cut, so we deal with those
+            # separately.
+
+            if isinstance(obj, generic.IndirectObject):
+                ref = obj.reference
+                if ref in seen_in_path:
+                    return
+                collected[ref].add(cur_path)
+                seen_in_path = seen_in_path.cons(ref)
+                obj = self(ref)
+                if not is_page_tree and ref in page_tree_objs:
+                    return
+            if isinstance(obj, generic.DictionaryObject):
+                for k, v in obj.items():
+                    # another hack to eliminate some spurious extra paths
+                    # that don't convey any useful information
+                    if k == '/Parent':
+                        continue
+
+                    _compute_paths_to_refs(
+                        v, cur_path.cons(k), seen_in_path,
+                        is_page_tree=is_page_tree or (
+                            cur_path.head == '/Root' and k == '/Pages'
+                            and cur_path.tail == misc.ConsList.empty()
+                        ),
+                        page_tree_objs=page_tree_objs
+                    )
+            elif isinstance(obj, generic.ArrayObject):
+                for ix, v in enumerate(obj):
+                    _compute_paths_to_refs(
+                        v, cur_path.cons(ix), seen_in_path,
+                        is_page_tree=is_page_tree,
+                        page_tree_objs=page_tree_objs
+                    )
+
+        def _collect_page_tree_refs(pages_obj):
+            for kid in pages_obj['/Kids']:
+                # should always be true, but hey
+                if isinstance(kid, generic.IndirectObject):
+                    yield kid.reference
+                kid = kid.get_object()
+                if kid.get('/Type', None) == '/Pages':
+                    yield from _collect_page_tree_refs(kid)
+
+        pages_ref = self.root.raw_get('/Pages')
+        page_tree_nodes = set()
+        page_tree_nodes.update(
+            _collect_page_tree_refs(pages_obj=pages_ref.get_object())
+        )
+        _compute_paths_to_refs(
+            self.trailer_view, misc.ConsList.empty(), misc.ConsList.empty(),
+            is_page_tree=False, page_tree_objs=page_tree_nodes
+        )
+
+        self._indirect_object_access_cache = {
+            ref: {RawPdfPath(*reversed(list(p))) for p in paths}
+            for ref, paths in collected.items()
+        }
