@@ -7,17 +7,23 @@ for the original license.
 
 import os
 import struct
-from hashlib import md5
 from io import BytesIO
-from typing import List, Union, Optional
+from typing import List, Union, Optional, Tuple
+
+from asn1crypto import x509
 
 from pyhanko.pdf_utils import generic
+from pyhanko.pdf_utils.crypt import (
+    SecurityHandler, StandardSecurityHandler,
+    PubKeySecurityHandler,
+)
 from pyhanko.pdf_utils.generic import pdf_name, pdf_string
 from pyhanko.pdf_utils.misc import (
     peek, PdfReadError, instance_test,
     PdfWriteError,
 )
 from pyhanko.pdf_utils.rw_common import PdfHandler
+from pyhanko import __version__
 
 
 __all__ = [
@@ -26,7 +32,7 @@ __all__ = [
 ]
 
 # TODO include version number
-VENDOR = 'pyhanko'
+VENDOR = 'pyHanko ' + __version__
 
 
 OBJSTREAM_FORBIDDEN = (generic.IndirectObject, generic.StreamObject)
@@ -107,16 +113,6 @@ class ObjectStream:
         return stream_object
 
 
-def _derive_key(base_key, idnum, generation):
-    # Ripped out of PyPDF2
-    # See § 7.6.2 in ISO 32000
-    pack1 = struct.pack("<i", idnum)[:3]
-    pack2 = struct.pack("<i", generation)[:2]
-    key = base_key + pack1 + pack2
-    md5_hash = md5(key).digest()
-    return md5_hash[:min(16, len(base_key) + 5)]
-
-
 def _contiguous_xref_chunks(position_dict):
     """
     Helper method to divide the XRef table (or stream) into contiguous chunks.
@@ -205,11 +201,9 @@ class XRefStream(generic.StreamObject):
             pdf_name('/Type'): pdf_name('/XRef'),
         })
 
-    def write_to_stream(self, stream, encryption_key):
+    def write_to_stream(self, stream, handler=None, container_ref=None):
         # the caller is responsible for making sure that the stream
         # is registered in the position dictionary
-        if encryption_key is not None:
-            raise ValueError('XRef streams cannot be encrypted')
 
         index = [0, 1]
         subsections = _contiguous_xref_chunks(self.position_dict)
@@ -299,6 +293,7 @@ class BasePdfFileWriter(PdfHandler):
         else:
             self._root = self.add_object(root)
 
+        self.security_handler: Optional[SecurityHandler] = None
         self._encrypt = self._encrypt_key = None
         self._document_id = document_id
         self.stream_xrefs = stream_xrefs
@@ -324,6 +319,11 @@ class BasePdfFileWriter(PdfHandler):
             self._info = info
         return info
 
+    @property
+    def document_id(self) -> Tuple[bytes, bytes]:
+        id_arr = self._document_id
+        return id_arr[0].original_bytes, id_arr[1].original_bytes
+
     def mark_update(self, obj_ref: Union[generic.Reference,
                                          generic.IndirectObject]):
         """
@@ -459,6 +459,10 @@ class BasePdfFileWriter(PdfHandler):
     def _write_header(self, stream):
         pass
 
+    def _assign_security_handler(self, sh: SecurityHandler):
+        self.security_handler = sh
+        self._encrypt = self.add_object(sh.as_pdf_object())
+
     def _write_objects(self, stream, object_position_dict):
         # deal with objects in object streams first
         for obj_stream in self.object_streams:
@@ -474,12 +478,14 @@ class BasePdfFileWriter(PdfHandler):
             generation, idnum = ix
             obj = self.objects[ix]
             object_position_dict[ix] = stream.tell()
-            stream.write(('%d %d obj' % (idnum, generation)).encode('ascii'))
-            if self._encrypt is not None and idnum != self._encrypt.idnum:
-                key = _derive_key(self._encrypt_key, idnum, generation)
+            stream.write(('%d %d obj\n' % (idnum, generation)).encode('ascii'))
+            if self.security_handler is not None \
+                    and idnum != self._encrypt.idnum:
+                handler = self.security_handler
             else:
-                key = None
-            obj.write_to_stream(stream, key)
+                handler = None
+            container_ref = generic.Reference(idnum, generation, self)
+            obj.write_to_stream(stream, handler, container_ref)
             stream.write(b'\nendobj\n')
 
     def _populate_trailer(self, trailer):
@@ -487,6 +493,8 @@ class BasePdfFileWriter(PdfHandler):
         trailer[pdf_name('/Root')] = self._root
         if self._info is not None:
             trailer[pdf_name('/Info')] = self._info
+        if self._encrypt is not None:
+            trailer[pdf_name('/Encrypt')] = self._encrypt
         # before doing anything else, we attempt to load the crypto-relevant
         # data, so that we can bail early if something's not right
         trailer[pdf_name('/ID')] = self._document_id
@@ -630,6 +638,12 @@ class BasePdfFileWriter(PdfHandler):
         Deep-copy an object into this writer, dealing with resolving indirect
         references in the process.
 
+        .. danger::
+            The table mapping indirect references in the input to indirect
+            references in the writer is not preserved between calls.
+            Concretely, this means that invoking :meth:`import_object` twice
+            on the same input reader may cause objects duplication.
+
         :param obj:
             The object to import.
         :param obj_stream:
@@ -807,7 +821,7 @@ class PageObject(generic.DictionaryObject):
 class PdfFileWriter(BasePdfFileWriter):
     """Class to write new PDF files."""
 
-    def __init__(self, stream_xrefs=True):
+    def __init__(self, stream_xrefs=True, init_page_tree=True):
         # root object
         root = generic.DictionaryObject({
             pdf_name("/Type"): pdf_name("/Catalog"),
@@ -824,13 +838,14 @@ class PdfFileWriter(BasePdfFileWriter):
 
         super().__init__(root, info, id_obj, stream_xrefs=stream_xrefs)
 
-        pages = generic.DictionaryObject({
-            pdf_name("/Type"): pdf_name("/Pages"),
-            pdf_name("/Count"): generic.NumberObject(0),
-            pdf_name("/Kids"): generic.ArrayObject(),
-        })
+        if init_page_tree:
+            pages = generic.DictionaryObject({
+                pdf_name("/Type"): pdf_name("/Pages"),
+                pdf_name("/Count"): generic.NumberObject(0),
+                pdf_name("/Kids"): generic.ArrayObject(),
+            })
 
-        root[pdf_name('/Pages')] = self.add_object(pages)
+            root[pdf_name('/Pages')] = self.add_object(pages)
 
     def _write_header(self, stream):
         major, minor = self.output_version
@@ -839,7 +854,83 @@ class PdfFileWriter(BasePdfFileWriter):
         # as binary (see § 7.5.2 in ISO 32000-1)
         stream.write(b'%\xc2\xa5\xc2\xb1\xc3\xab\n')
 
-    # I can't be arsed to actually implement encrypt() for newly written PDFs,
-    # since all security handlers specified in the 1.7 standard are insecure as
-    # hell. I'm going to keep the RC4 functionality in the incremental writer,
-    # since we still want to be able to sign old PDF files.
+    def encrypt(self, owner_pass, user_pass=None):
+        """
+        Mark this document to be encrypted with PDF 2.0 encryption (AES-256).
+
+        .. caution::
+            While pyHanko supports legacy PDF encryption as well, the API
+            to create new documents using outdated encryption is left
+            largely undocumented on purpose to discourage its use.
+
+            This caveat does *not* apply to incremental updates added to
+            existing documents.
+
+        .. danger::
+            The PDF 2.0 standard mandates AES-256 in CBC mode, and also includes
+            12 bytes of known plaintext by design. This implies that a
+            sufficiently knowledgeable attacker can inject arbitrary content
+            into your encrypted files without knowledge of the password.
+
+            Adding a digital signature to the encrypted document is **not**
+            a foolproof way to deal with this either, since most viewers will
+            still allow the document to be opened before signatures are
+            validated, and therefore end users are still exposed to potentially
+            malicious content.
+
+            Until the standard supports authenticated encryption schemes, you
+            should **never** rely on its encryption provisions if tampering
+            is a concern.
+
+
+        :param owner_pass:
+            The desired owner password.
+        :param user_pass:
+            The desired user password (defaults to the owner password
+            if not specified)
+        """
+        self.output_version = (2, 0)
+        sh = StandardSecurityHandler.build_from_pw(owner_pass, user_pass)
+        self._assign_security_handler(sh)
+
+    def encrypt_pubkey(self, recipients: List[x509.Certificate]):
+        """
+        Mark this document to be encrypted with PDF 2.0 public key encryption.
+        The certificates passed in should be RSA certificates.
+
+        PyHanko defaults to AES-256 to encrypt the actual file contents.
+        The seed used to derive the file encryption key is also encrypted
+        using AES-256 and bundled in a CMS EnvelopedData object.
+        The envelope key is then encrypted separately for each recipient, using
+        their respective public keys.
+
+        .. caution::
+            The caveats for :meth:`encrypt` also apply here.
+
+        :param recipients:
+            Certificates of the recipients that should be able to decrypt
+            the document.
+        """
+        self.output_version = (2, 0)
+        sh = PubKeySecurityHandler.build_from_certs(recipients)
+        self._assign_security_handler(sh)
+
+
+def copy_into_new_writer(input_handler: PdfHandler) -> PdfFileWriter:
+    w = PdfFileWriter(init_page_tree=False)
+    input_root_ref = input_handler.root_ref
+    output_root_ref = w.root_ref
+    # call _import_object in such a way that we translate the input handler's
+    # root to the new writer's root.
+    # From a technical PoV this doesn't matter, but it makes the output file
+    # somewhat "cleaner" (i.e. it doesn't leave an orphaned document catalog
+    # cluttering up the file)
+    new_root_dict = w._import_object(
+        input_handler.root, reference_map={input_root_ref: output_root_ref},
+        obj_stream=None
+    )
+    # override the old root ref
+    ix = (output_root_ref.generation, output_root_ref.idnum)
+    w.objects[ix] = new_root_dict
+
+    return w
