@@ -61,12 +61,15 @@ from enum import unique
 from io import BytesIO
 from typing import (
     Iterable, Optional, Set, Tuple, Generator, TypeVar, Dict,
-    List, Callable, Union,
+    List, Callable, Union
 )
 
 from pyhanko.pdf_utils.generic import Reference, PdfObject
 from pyhanko.pdf_utils.misc import OrderedEnum
-from pyhanko.pdf_utils.reader import HistoricalResolver, PdfFileReader
+from pyhanko.pdf_utils.reader import (
+    HistoricalResolver, PdfFileReader,
+    RawPdfPath,
+)
 from pyhanko.pdf_utils import generic, misc
 from pyhanko.sign.fields import FieldMDPSpec, MDPPerm
 
@@ -153,6 +156,25 @@ class SuspiciousModification(ValueError):
     pass
 
 
+@dataclass(frozen=True)
+class ReferenceUpdate:
+
+    updated_ref: Reference
+    """
+    Reference that was (potentially) updated.
+    """
+
+    # TODO document
+    paths_checked: Optional[Union[RawPdfPath, Iterable[RawPdfPath]]] \
+        = None
+
+    blanket_approve: bool = False
+
+    @classmethod
+    def curry_ref(cls, **kwargs):
+        return lambda ref: cls(updated_ref=ref, **kwargs)
+
+
 class QualifiedWhitelistRule:
     """
     Abstract base class for a whitelisting rule that outputs references together
@@ -163,7 +185,7 @@ class QualifiedWhitelistRule:
     """
 
     def apply_qualified(self, old: HistoricalResolver, new: HistoricalResolver)\
-            -> Iterable[Tuple[ModificationLevel, Reference]]:
+            -> Iterable[Tuple[ModificationLevel, ReferenceUpdate]]:
         """
         Apply the rule to the changes between two revisions.
 
@@ -186,7 +208,7 @@ class WhitelistRule:
     """
 
     def apply(self, old: HistoricalResolver, new: HistoricalResolver) \
-            -> Iterable[Reference]:
+            -> Iterable[ReferenceUpdate]:
         """
         Apply the rule to the changes between two revisions.
 
@@ -218,22 +240,23 @@ class _WrappingQualifiedWhitelistRule(QualifiedWhitelistRule):
         self.level = level
 
     def apply_qualified(self, old: HistoricalResolver, new: HistoricalResolver)\
-            -> Iterable[Tuple[ModificationLevel, Reference]]:
+            -> Iterable[Tuple[ModificationLevel, ReferenceUpdate]]:
         for ref in self.rule.apply(old, new):
             yield self.level, ref
 
 
 R = TypeVar('R')
+X = TypeVar('X')
 
 
 def qualify(level: ModificationLevel,
-            rule_result: Generator[Reference, None, R],
-            transform=lambda x: x)\
-        -> Generator[Tuple[ModificationLevel, Reference], None, R]:
+            rule_result: Generator[X, None, R],
+            transform: Callable[[X], ReferenceUpdate] = lambda x: x)\
+        -> Generator[Tuple[ModificationLevel, ReferenceUpdate], None, R]:
     """
     This is a helper function for rule implementors.
-    It attaches a fixed modification level to an existing reference generator,
-    respecting the original generator's return value (if relevant).
+    It attaches a fixed modification level to an existing reference update
+    generator, respecting the original generator's return value (if relevant).
 
     A prototypical use would be of the following form:
 
@@ -257,7 +280,7 @@ def qualify(level: ModificationLevel,
             )
 
     Provided that ``some_generator_function`` yields
-    :class:`~.generic.Reference` objects, the yield type of the resulting
+    :class:`~.generic.ReferenceUpdate` objects, the yield type of the resulting
     generator will be tuples of the form ``(level, ref)``.
 
     :param level:
@@ -277,7 +300,8 @@ def qualify(level: ModificationLevel,
     )
 
 
-def _safe_whitelist(old: HistoricalResolver, old_ref, new_ref):
+def _safe_whitelist(old: HistoricalResolver, old_ref, new_ref) \
+        -> Generator[Reference, None, None]:
     if old_ref:
         _assert_not_stream(old_ref.get_object())
 
@@ -298,7 +322,7 @@ class DocInfoRule(WhitelistRule):
     """
 
     def apply(self, old: HistoricalResolver, new: HistoricalResolver) \
-            -> Iterable[Reference]:
+            -> Iterable[ReferenceUpdate]:
         # updates to /Info are always OK (and must be through indirect objects)
         # Removing the /Info dictionary is no big deal, since most readers
         # will fall back to older revisions regardless
@@ -310,7 +334,29 @@ class DocInfoRule(WhitelistRule):
         old_info = old.trailer_view.get_value_as_reference(
             '/Info', optional=True
         )
-        yield from _safe_whitelist(old, old_info, new_info)
+        path = RawPdfPath('/Info')
+        yield from map(
+            ReferenceUpdate.curry_ref(paths_checked=path),
+            _safe_whitelist(old, old_info, new_info)
+        )
+
+
+def _validate_dss_substructure(old: HistoricalResolver, new: HistoricalResolver,
+                               new_dict, der_stream_keys, is_vri):
+    for der_obj_type in der_stream_keys:
+        try:
+            value = new_dict.raw_get(der_obj_type)
+        except KeyError:
+            continue
+        if not isinstance(value.get_object(), generic.ArrayObject):
+            raise SuspiciousModification(
+                f"Expected array at {'VRI' if is_vri else 'DSS'} "
+                f"key {der_obj_type}."
+            )
+
+        yield from map(ReferenceUpdate, new.collect_dependencies(
+            value, since_revision=old.revision + 1
+        ))
 
 
 class DSSCompareRule(WhitelistRule):
@@ -325,11 +371,13 @@ class DSSCompareRule(WhitelistRule):
     """
 
     def apply(self, old: HistoricalResolver, new: HistoricalResolver)\
-            -> Iterable[Reference]:
+            -> Iterable[ReferenceUpdate]:
         # TODO refactor these into less ad-hoc rules
 
-        old_dss, new_dss = yield from _compare_key_refs(
-            '/DSS', old, old.root, new.root
+        dss_path = RawPdfPath('/Root', '/DSS')
+        old_dss, new_dss = yield from misc.map_with_return(
+            _compare_key_refs('/DSS', old, old.root, new.root),
+            ReferenceUpdate.curry_ref(paths_checked=dss_path)
         )
         if new_dss is None:
             return
@@ -351,24 +399,19 @@ class DSSCompareRule(WhitelistRule):
                 f"Unexpected keys in DSS: {dss_keys - dss_expected_keys}."
             )
 
-        for der_obj_type in dss_der_stream_keys:
-            try:
-                value = new_dss.raw_get(der_obj_type)
-            except KeyError:
-                continue
-            if not isinstance(value.get_object(), generic.ArrayObject):
-                raise SuspiciousModification(
-                    f"Expected array at DSS key {der_obj_type}"
-                )
-
-            yield from new.collect_dependencies(
-                value, since_revision=old.revision + 1
-            )
+        yield from _validate_dss_substructure(
+            old, new, new_dss, dss_der_stream_keys, is_vri=False
+        )
 
         # check that the /VRI dictionary still contains all old keys, unchanged.
-        old_vri, new_vri = yield from _compare_key_refs(
-            '/VRI', old, old_dss, new_dss,
+        vri_path = RawPdfPath('/Root', '/DSS', '/VRI')
+        old_vri, new_vri = yield from misc.map_with_return(
+            _compare_key_refs(
+                '/VRI', old, old_dss, new_dss,
+            ), ReferenceUpdate.curry_ref(paths_checked=vri_path)
+
         )
+
         if old_vri is None:
             old_vri = generic.DictionaryObject()
 
@@ -403,7 +446,7 @@ class DSSCompareRule(WhitelistRule):
             new_vri_dict = new_vri.raw_get(key)
             if isinstance(new_vri_dict, generic.IndirectObject) \
                     and old.is_ref_available(new_vri_dict.reference):
-                yield new_vri_dict.reference
+                yield ReferenceUpdate(new_vri_dict.reference)
                 new_vri_dict = new_vri_dict.get_object()
             _assert_not_stream(new_vri_dict)
             if not isinstance(new_vri_dict, generic.DictionaryObject):
@@ -417,18 +460,9 @@ class DSSCompareRule(WhitelistRule):
                     "Unexpected keys in VRI dictionary: "
                     f"{new_vri_value_keys - vri_expected_keys}."
                 )
-            for der_obj_type in vri_der_stream_keys:
-                try:
-                    value = new_vri_dict.raw_get(der_obj_type)
-                except KeyError:
-                    continue
-                if not isinstance(value.get_object(), generic.ArrayObject):
-                    raise SuspiciousModification(
-                        f"Expected array at VRI key {der_obj_type}"
-                    )
-                yield from new.collect_dependencies(
-                    value, since_revision=old.revision + 1
-                )
+            yield from _validate_dss_substructure(
+                old, new, new_vri_dict, vri_der_stream_keys, is_vri=True
+            )
 
             # /TS is also a DER stream
             try:
@@ -436,7 +470,7 @@ class DSSCompareRule(WhitelistRule):
                     '/TS', optional=True
                 )
                 if ts_ref is not None and old.is_ref_available(ts_ref):
-                    yield ts_ref
+                    yield ReferenceUpdate(ts_ref)
             except misc.IndirectObjectExpected:
                 pass
 
@@ -463,6 +497,12 @@ class FieldComparisonSpec:
     A reference to the field's dictionary in the new revision, if present.
     """
 
+    old_canonical_path: Optional[RawPdfPath]
+    """
+    Path from the trailer through the AcroForm structure to this field (in the
+    older revision). If the field is new, set to ``None``.
+    """
+
     @property
     def old_field(self) -> Optional[generic.DictionaryObject]:
         """
@@ -491,6 +531,46 @@ class FieldComparisonSpec:
         assert isinstance(field, generic.DictionaryObject)
         return field
 
+    def old_annotation_paths(self):
+        # collect path(s) through which this field is used as an annotation
+        # the clean way to accomplish this would be to follow /P
+        # and go from there, but /P is optional, so we have to get a little
+        # creative.
+        old_field_ref = self.old_field_ref
+        if old_field_ref is None:
+            return set()
+
+        old = self.old_field_ref.get_pdf_handler()
+        assert isinstance(old, HistoricalResolver)
+
+        all_paths = old._get_usages_of_ref(old_field_ref)
+
+        def _path_ok(pdf_path: RawPdfPath):
+            # check if the path looks like a path to an annotation on a page
+
+            # .Root.Pages.Kids[0].Annots[0] is the shortest you can get,
+            # so 6 nodes is the minimum
+            if len(pdf_path) < 6:
+                return False
+            fst, snd, *rest = pdf_path.path
+            if fst != '/Root' or snd != '/Pages':
+                return False
+
+            # there should be one or more elements of the form /Kids[i] now
+            descended = False
+            nxt, nxt_ix, *rest = rest
+            while nxt == '/Kids' and isinstance(nxt_ix, int):
+                descended = True
+                nxt, nxt_ix, *rest = rest
+
+            # rest should be nothing and nxt should be /Annots
+            return (
+                descended and not rest and nxt == '/Annots'
+                and isinstance(nxt_ix, int)
+            )
+
+        return {p for p in all_paths if _path_ok(p)}
+
 
 @dataclass(frozen=True)
 class FieldComparisonContext:
@@ -515,7 +595,7 @@ class FieldComparisonContext:
 
 
 @dataclass(frozen=True)
-class FormUpdate:
+class FormUpdate(ReferenceUpdate):
     """
     Container for a reference together with (optional) metadata.
 
@@ -524,12 +604,7 @@ class FormUpdate:
     is locked by the FieldMDP policy currently in force.
     """
 
-    updated_ref: generic.Reference
-    """
-    Reference that was (potentially) updated.
-    """
-
-    field_name: Optional[str]
+    field_name: Optional[str] = None
     """
     The relevant field's fully qualified name, or ``None`` if there's either
     no obvious associated field, or if there are multiple reasonable candidates.
@@ -541,10 +616,6 @@ class FormUpdate:
     This is only relevant if :attr:`field_name` is not ``None``.
     """
 
-    @staticmethod
-    def curry_ref(**kwargs):
-        return lambda ref: FormUpdate(updated_ref=ref, **kwargs)
-
 
 class FieldMDPRule:
     """
@@ -632,10 +703,11 @@ class SigFieldCreationRule(FieldMDPRule):
             # Since LTA updates should arguably not trigger field locks either
             # (relevant for FieldMDP settings that use /All or /Exclude),
             # we pass valid_when_locked=True on these updates
-            yield ModificationLevel.LTA_UPDATES, FormUpdate(
-                updated_ref=sigfield_ref, field_name=fq_name,
-                valid_when_locked=True
-            )
+            if context.old.is_ref_available(sigfield_ref):
+                yield ModificationLevel.LTA_UPDATES, FormUpdate(
+                    updated_ref=sigfield_ref, field_name=fq_name,
+                    valid_when_locked=True
+                )
             sigfield = sigfield_ref.get_object()
             # checked by field listing routine already
             assert isinstance(sigfield, generic.DictionaryObject)
@@ -801,9 +873,15 @@ class SigFieldModificationRule(BaseFieldModificationRule):
             # and also register whether the changes made would be
             # permissible even when the field is locked.
             valid_when_locked = self.compare_fields(spec)
+
+            # these are the paths where we expect to encounter this signature
+            # field
+            paths = spec.old_annotation_paths()
+            paths.add(spec.old_canonical_path)
             field_ref_update = FormUpdate(
                 updated_ref=spec.new_field_ref, field_name=fq_name,
-                valid_when_locked=valid_when_locked
+                valid_when_locked=valid_when_locked,
+                paths_checked=paths
             )
 
             if not previously_signed and now_signed:
@@ -915,11 +993,15 @@ class GenericFieldModificationRule(BaseFieldModificationRule):
             return
 
         valid_when_locked = self.compare_fields(spec)
+
+        # these are the paths where we expect the form field to be referred to
+        paths = spec.old_annotation_paths()
+        paths.add(spec.old_canonical_path)
         yield (
             ModificationLevel.FORM_FILLING,
             FormUpdate(
                 updated_ref=spec.new_field_ref, field_name=fq_name,
-                valid_when_locked=valid_when_locked
+                valid_when_locked=valid_when_locked, paths_checked=paths
             )
         )
         old_field = spec.old_field
@@ -955,6 +1037,9 @@ class GenericFieldModificationRule(BaseFieldModificationRule):
             )
 
 
+ACROFORM_EXEMPT_STRICT_COMPARISON = {'/Fields', '/DR'}
+
+
 class FormUpdatingRule:
     """
     Special whitelisting rule that validates changes to the form attached to
@@ -986,7 +1071,7 @@ class FormUpdatingRule:
         self.field_rules = field_rules
         self.ignored_acroform_keys = (
             ignored_acroform_keys if ignored_acroform_keys is not None
-            else {'/Fields'}
+            else ACROFORM_EXEMPT_STRICT_COMPARISON
         )
 
     def apply(self, old: HistoricalResolver, new: HistoricalResolver)\
@@ -1000,17 +1085,15 @@ class FormUpdatingRule:
             The newer revision to be vetted.
         """
 
-        def _emit_ref(_ref):
-            return (
-                ModificationLevel.LTA_UPDATES,
-                FormUpdate(updated_ref=_ref, field_name=None)
-            )
-
-        old_acroform, new_acroform = yield from misc.map_with_return(
+        acroform_path = RawPdfPath('/Root', '/AcroForm')
+        old_acroform, new_acroform = yield from qualify(
+            ModificationLevel.LTA_UPDATES,
             _compare_key_refs(
                 '/AcroForm', old, old.root, new.root
             ),
-            _emit_ref
+            transform=FormUpdate.curry_ref(
+                field_name=None, paths_checked=acroform_path
+            )
         )
 
         # first, compare the entries that aren't /Fields
@@ -1022,18 +1105,37 @@ class FormUpdatingRule:
         # This is fine: the _list_fields logic checks that it really contains
         # stuff that looks like form fields, and other rules are responsible
         # for vetting the creation of other form fields anyway.
-        yield from misc.map_with_return(
+        fields_path = acroform_path + '/Fields'
+        old_fields, new_fields = yield from qualify(
+            ModificationLevel.LTA_UPDATES,
             _compare_key_refs('/Fields', old, old_acroform, new_acroform),
-            _emit_ref
+            transform=FormUpdate.curry_ref(
+                field_name=None, paths_checked=fields_path
+            )
         )
-        try:
-            old_fields = old_acroform['/Fields']
-            new_fields = new_acroform['/Fields']
-        except KeyError:  # pragma: nocover
-            raise misc.PdfReadError("Could not read /Fields in form")
+
+        # we also need to deal with the default resource dict, since
+        # Acrobat / Adobe Reader sometimes mess with it
+        old_dr, new_dr = yield from qualify(
+            ModificationLevel.FORM_FILLING,
+            _compare_key_refs('/DR', old, old_acroform, new_acroform),
+            transform=FormUpdate.curry_ref(
+                field_name=None, paths_checked=acroform_path + '/DR'
+            )
+        )
+        if new_dr is not None:
+            dr_deps = new.collect_dependencies(
+                new_dr, since_revision=old.revision + 1
+            )
+            yield from qualify(
+                ModificationLevel.FORM_FILLING, misc._as_gen(dr_deps),
+                transform=FormUpdate.curry_ref(field_name=None)
+            )
 
         context = FieldComparisonContext(
-            field_specs=dict(_list_fields(old_fields, new_fields)),
+            field_specs=dict(
+                _list_fields(old_fields, new_fields, old_path=fields_path)
+            ),
             old=old, new=new
         )
 
@@ -1042,7 +1144,7 @@ class FormUpdatingRule:
 
 
 ROOT_EXEMPT_STRICT_COMPARISON = {
-    '/AcroForm', '/DSS', '/Extensions', '/Metadata', '/MarkInfo'
+    '/AcroForm', '/DSS', '/Extensions', '/Metadata', '/MarkInfo', '/Version'
 }
 
 
@@ -1081,7 +1183,14 @@ class CatalogModificationRule(QualifiedWhitelistRule):
         #  - /MarkInfo: if it's an indirect reference (probably not) we can
         #    whitelist it if the key set makes sense. TODO do this
         #  - /DSS, /AcroForm and /Metadata are dealt with by other rules.
-        yield ModificationLevel.LTA_UPDATES, new.root_ref
+        yield ModificationLevel.LTA_UPDATES, ReferenceUpdate(
+            new.root_ref, paths_checked=RawPdfPath('/Root'),
+            # Things like /Data in a MDP policy can point to root
+            # and since we checked with _compare_dicts, doing a blanket
+            # approval is much easier than figuring out all the ways
+            # in which /Root can be cross-referenced.
+            blanket_approve=True
+        )
 
 
 class MetadataUpdateRule(WhitelistRule):
@@ -1149,7 +1258,7 @@ class MetadataUpdateRule(WhitelistRule):
             )
 
     def apply(self, old: HistoricalResolver, new: HistoricalResolver) \
-            -> Iterable[Reference]:
+            -> Iterable[ReferenceUpdate]:
 
         # /Metadata points to a stream, so we have to be careful allowing
         # object overrides!
@@ -1183,7 +1292,10 @@ class MetadataUpdateRule(WhitelistRule):
             and not self.always_refuse_stream_override
         )
         if same_ref_ok or old.is_ref_available(new_metadata_ref):
-            yield new_metadata_ref
+            yield ReferenceUpdate(
+                new_metadata_ref,
+                paths_checked=RawPdfPath('/Root', '/Metadata')
+            )
 
 
 class ObjectStreamRule(WhitelistRule):
@@ -1200,7 +1312,7 @@ class ObjectStreamRule(WhitelistRule):
         # object streams are OK, but overriding object streams is not.
         for objstream_ref in new.object_streams_used():
             if old.is_ref_available(objstream_ref):
-                yield objstream_ref
+                yield ReferenceUpdate(objstream_ref)
 
 
 class XrefStreamRule(WhitelistRule):
@@ -1213,12 +1325,12 @@ class XrefStreamRule(WhitelistRule):
         xref_start, _ = new.reader.xrefs.get_xref_container_info(new.revision)
         if isinstance(xref_start, generic.Reference) \
                 and old.is_ref_available(xref_start):
-            yield xref_start
+            yield ReferenceUpdate(xref_start)
 
 
 def _list_fields(old_fields: generic.PdfObject, new_fields: generic.PdfObject,
-                 parent_name="",
-                 inherited_ft=None) -> Dict[str, FieldComparisonSpec]:
+                 old_path: RawPdfPath, parent_name="", inherited_ft=None) \
+        -> Dict[str, FieldComparisonSpec]:
     """
     Recursively construct a list of field names, together with their
     "incarnations" in either revision.
@@ -1229,7 +1341,7 @@ def _list_fields(old_fields: generic.PdfObject, new_fields: generic.PdfObject,
             raise exc("Field list is not an array.")
         names_seen = set()
 
-        for field_ref in lst:
+        for ix, field_ref in enumerate(lst):
             if not isinstance(field_ref, generic.IndirectObject):
                 raise exc("Fields must be indirect objects")
 
@@ -1259,7 +1371,7 @@ def _list_fields(old_fields: generic.PdfObject, new_fields: generic.PdfObject,
                 kids = field["/Kids"]
             except KeyError:
                 kids = generic.ArrayObject()
-            yield fq_name, (field_type, field_ref.reference, kids)
+            yield fq_name, (field_type, field_ref.reference, kids, ix)
 
     old_fields_by_name = dict(_make_list(old_fields, misc.PdfReadError))
     new_fields_by_name = dict(_make_list(new_fields, SuspiciousModification))
@@ -1270,14 +1382,15 @@ def _list_fields(old_fields: generic.PdfObject, new_fields: generic.PdfObject,
 
     for field_name in names:
         try:
-            old_field_type, old_field_ref, old_kids = \
+            old_field_type, old_field_ref, old_kids, field_index = \
                 old_fields_by_name[field_name]
         except KeyError:
             old_field_type = old_field_ref = None
             old_kids = generic.ArrayObject()
+            field_index = None
 
         try:
-            new_field_type, new_field_ref, new_kids = \
+            new_field_type, new_field_ref, new_kids, _ = \
                 new_fields_by_name[field_name]
         except KeyError:
             new_field_type = new_field_ref = None
@@ -1289,21 +1402,29 @@ def _list_fields(old_fields: generic.PdfObject, new_fields: generic.PdfObject,
                     f"Update changed field type of {field_name}"
                 )
         common_ft = old_field_type or new_field_type
+        if field_index is not None and old_path is not None:
+            field_path = old_path + field_index
+        else:
+            field_path = None
         yield field_name, FieldComparisonSpec(
             field_type=common_ft,
-            old_field_ref=old_field_ref, new_field_ref=new_field_ref
+            old_field_ref=old_field_ref, new_field_ref=new_field_ref,
+            old_canonical_path=field_path
         )
 
         # recursively descend into /Kids if necessary
         if old_kids or new_kids:
             yield from _list_fields(
-                old_kids, new_kids, field_name, inherited_ft=common_ft
+                old_kids, new_kids, parent_name=field_name, old_path=(
+                    field_path + '/Kids' if field_path is not None else None
+                ),
+                inherited_ft=common_ft,
             )
 
 
 def _allow_appearance_update(old_field, new_field, old: HistoricalResolver,
                              new: HistoricalResolver) \
-        -> Generator[generic.Reference, None, None]:
+        -> Generator[Reference, None, None]:
 
     old_ap_val, new_ap_val = yield from _compare_key_refs(
         '/AP', old, old_field, new_field
@@ -1445,17 +1566,29 @@ def _walk_page_tree_annots(old_page_root, new_page_root,
                 field_name = field_name_dict[next(iter(added_annots))]
 
             # Make sure the page dictionaries are the same, so that we
-            #  can safely clear them for modification
+            #  can safely clear them for modification across ALL paths
             #  (not necessary if both /Annots entries are indirect references,
             #   but adding even more cases is pushing things)
             _compare_dicts(old_kid, new_kid, {'/Annots'})
+            # Page objects are often referenced from all sorts of places in the
+            # file, and attempting to check all possible paths would probably
+            # create more problems than it solves.
             yield FormUpdate(
                 updated_ref=new_kid_ref, field_name=field_name,
-                valid_when_locked=valid_when_locked and field_name is not None
+                valid_when_locked=valid_when_locked and field_name is not None,
+                blanket_approve=True
             )
             if new_annots_ref:
                 # current /Annots entry is an indirect reference
 
+                # collect paths to this page and append /Annots
+                #  (recall: old_kid_ref and new_kid_ref should be the same
+                #   anyhow)
+                paths_to_annots = {
+                    path + '/Annots'
+                    for path in old._get_usages_of_ref(old_kid_ref)
+                }
+
                 # If the equality check fails,
                 # either the /Annots array got reassigned to another
                 # object ID, or it was moved from a direct object to an
@@ -1468,7 +1601,8 @@ def _walk_page_tree_annots(old_page_root, new_page_root,
                         updated_ref=new_annots_ref, field_name=field_name,
                         valid_when_locked=(
                             valid_when_locked and field_name is not None
-                        )
+                        ),
+                        paths_checked=paths_to_annots
                     )
 
 
@@ -1564,7 +1698,7 @@ def _compare_key_refs(key, old: HistoricalResolver,
             raise SuspiciousModification(
                 f"Key {key} was deleted from dictionary"
             )
-        return old_value, None  # nothing to do
+        return None, None  # nothing to do
 
     if new_value_ref is not None:
         yield from _safe_whitelist(old, old_value_ref, new_value_ref)
@@ -1702,9 +1836,52 @@ class StandardDiffPolicy(DiffPolicy):
 
         explained = defaultdict(set)
 
+        # prepare LUT for refs that are used multiple times in the old revision
+        # (this is a very expensive operation, since it reads all objects in
+        #  the signed revision)
+        def _init_multi_lut():
+            old._load_reverse_xref_cache()
+            for _ref in new_xrefs:
+                usages = old._get_usages_of_ref(_ref)
+                if usages:
+                    yield _ref, (ModificationLevel.NONE, set(usages))
+
+        # This table records all the overridden refs that already existed
+        # in the old revision, together with the different ways they can be
+        # reached from the document trailer.
+        # Unlike fresh refs, these need to be cleared together with the paths
+        # through which they are accessed.
+        old_usages_to_clear = dict(_init_multi_lut())
+
+        def ingest_ref(_level: ModificationLevel, _upd: ReferenceUpdate):
+            ref = _upd.updated_ref
+            try:
+                current_max_level, usages = old_usages_to_clear[ref]
+                if _upd.blanket_approve:
+                    # approve all usages at once
+                    usages = set()
+                else:
+                    # remove the paths that have just been cleared from
+                    # the checklist
+                    paths_checked = _upd.paths_checked or ()
+                    if isinstance(paths_checked, RawPdfPath):
+                        # single path
+                        paths_checked = paths_checked,
+                    usages.difference_update(paths_checked)
+                # bump the modification level for this reference if necessary
+                _level = max(current_max_level, _level)
+                old_usages_to_clear[ref] = _level, usages
+                if usages:
+                    # not all paths/usages have been cleared, so we can't
+                    # approve the reference yet
+                    return
+            except KeyError:
+                pass
+            explained[_level].add(ref)
+
         for rule in self.global_rules:
-            for level, ref in rule.apply_qualified(old, new):
-                explained[level].add(ref)
+            for level, upd in rule.apply_qualified(old, new):
+                ingest_ref(level, upd)
 
         changed_form_fields = set()
 
@@ -1716,7 +1893,7 @@ class StandardDiffPolicy(DiffPolicy):
                        and field_mdp_spec.is_locked(fq_name)
 
             for level, fu in form_changes:
-                explained[level].add(fu.updated_ref)
+                ingest_ref(level, fu)
                 field_name = fu.field_name
                 if field_name is not None and not fu.valid_when_locked:
                     if is_locked(field_name):
@@ -1743,10 +1920,27 @@ class StandardDiffPolicy(DiffPolicy):
                 "Unexplained xrefs in revision %d:\n%s",
                 new.revision, msg
             )
-            raise SuspiciousModification(
+            unexplained_overrides = [
+                f" - {repr(ref)} is also used at "
+                f"{', '.join(str(p) for p in paths_remaining)} in the prior "
+                f"revision."
+                for ref, (_, paths_remaining) in old_usages_to_clear.items()
+                if paths_remaining
+            ]
+            err_msg = (
                 f"There are unexplained xrefs in revision {new.revision}: "
                 f"{', '.join(repr(x) for x in unexplained_annot)}."
             )
+            if unexplained_overrides:
+                unchecked_paths_msg = (
+                    f"Some objects from revision {old.revision} were replaced "
+                    f"in revision {new.revision} without precise "
+                    "justification:\n" + '\n'.join(unexplained_overrides)
+                )
+                err_msg = "%s\n%s" % (err_msg, unchecked_paths_msg)
+                logger.debug(unchecked_paths_msg)
+
+            raise SuspiciousModification(err_msg)
         elif unexplained_formfill:
             level = ModificationLevel.ANNOTATIONS
         elif unexplained_lta:
